/*++

Copyright (c) 2004 Alexander A. Telyatnikov (Alter)

Module Name:
    DbgPrintLog.cpp

Abstract:

Authors:
    Alexander A. Telyatnikov (Alter)
    Max Sivkov

Environment:
    User mode only

Notes:

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Revision History:
    Initial code                 - Alter
    Initial Service mode support - Max

--*/

#include <windows.h>
#include <stdio.h>
#include "..\Driver\DbgPrnHk.h"
#include "SvcManLib.h"
#include "Privilege.h"
#include "flusher.h"

#define NT_DbgPrnHk_Client_SVC_NAME     "DbgPrintLog"
#define NT_DbgPrnHk_Reg_CmdLineW        L"CmdLine"
#define NT_DbgPrnHk_Reg_WDirW           L"WorkingDirectory"

#define print_log  printf

typedef struct _DBG_OUTPUT_DEBUG_STRING_BUFFER {
    ULONG ProcessId;
    UCHAR Msg[4096-sizeof(ULONG)];
} DBG_OUTPUT_DEBUG_STRING_BUFFER, *PDBG_OUTPUT_DEBUG_STRING_BUFFER;

void
WINAPI
service_main(
    DWORD dwArgc,
    LPTSTR *lpszArgv);

VOID AddToMessageLog(LPTSTR lpszMsg);

BOOL ReportStatusToSCMgr(DWORD dwCurrentState,
                         DWORD dwWin32ExitCode,
                         DWORD dwWaitHint);

SERVICE_TABLE_ENTRY dispatchTable[] =
{
    { NT_DbgPrnHk_Client_SVC_NAME, (LPSERVICE_MAIN_FUNCTION)service_main },
    { NULL, NULL }
};

PWCHAR LogFile = L"DbgPrint.log";  // default log-file name
WCHAR LogFileN[MAX_PATH*2];
WCHAR LogFileSuf[MAX_PATH];

BOOLEAN esc_key       = FALSE;
BOOLEAN pause_key     = FALSE;
BOOLEAN next_log      = FALSE;
BOOLEAN flush_file    = FALSE;
BOOLEAN sys_account   = FALSE;

BOOLEAN log_pid       = FALSE;
BOOLEAN log_tid       = FALSE;
BOOLEAN log_mode      = FALSE;
BOOLEAN log_irql      = FALSE;
BOOLEAN log_time_ex   = FALSE;
BOOLEAN log_time_rel  = FALSE;
BOOLEAN log_time_date = FALSE;
BOOLEAN log_time_time = TRUE;

BOOLEAN skip_kernel   = FALSE;
BOOLEAN skip_umode    = FALSE;

BOOLEAN stdout_mode   = FALSE;
BOOLEAN sync_mode     = FALSE;
BOOLEAN stdout_copy   = FALSE;
BOOLEAN restart_drv   = FALSE;

BOOLEAN install_svc   = FALSE;
BOOLEAN install_drv   = FALSE;
BOOLEAN deinstall     = FALSE;

ULONG   drv_mode      = SERVICE_DEMAND_START;
ULONG   cli_mode      = SERVICE_AUTO_START;

PDbgPrnHk_GetMessages_USER_OUT MsgBuffer = NULL;

// dbg messages capture buffers for user-mode
// (via OutputDebugString)
// '2' suffix  - buffer/counter for reading from get_messages()
// '2a' suffix - buffer/counter for writing from get_ODS_messages()
HANDLE hODS_lock = NULL;
ULONG MsgCount2 = 0;
ULONG MsgCount2a = 0;
ULONG MaxLogs = -1;
PDbgPrnHk_GetMessages_USER_OUT MsgBuffer2 = NULL;
PDbgPrnHk_GetMessages_USER_OUT MsgBuffer2a = NULL;

// settings
ULONG BufferSize = 1024*1024;

ULONG MaxLogSize = 1024*1024*128; //128Mb

ULONG Start_i = 0;

ULONG MaxSleepCount = 5;

HANDLE StdIn  = NULL;
HANDLE StdOut = NULL;

SERVICE_STATUS          ssStatus;       // current status of the service
SERVICE_STATUS_HANDLE   sshStatusHandle;

WCHAR UserName[MAX_PATH];

BOOLEAN
IsSystemAccount()
{
    WCHAR UserName[MAX_PATH];
    ULONG len = MAX_PATH-1;

    if(!GetUserNameW(UserName, &len)) {
        UserName[0] = 0;
        return FALSE;
    }
    UserName[len] = 0;
    if(!wcsicmp(UserName, L"System")) {
        return TRUE;
    }
    return FALSE;
} // end IsSystemAccount()

VOID
TeeOutput(
    HANDLE h,
    PCHAR Buffer,
    ULONG Length
    )
{
    WriteFileBuffer(h, Buffer, Length);
    if(!stdout_copy || stdout_mode) {
        return;
    }
    WriteConsole(StdOut, Buffer, Length, &Length, NULL);
} // end TeeOutput()

/*
 Catch keyboard messages
*/
DWORD
__stdcall
get_kbd(PVOID ctx)
{
    ULONG key = 0;
    ULONG tmp;
    INPUT_RECORD inp;

    if(!StdIn)
        return -1;
    while(!esc_key) {
        if(GetNumberOfConsoleInputEvents(StdIn, &key) && key) {
            key = ReadConsoleInput(StdIn, &inp, 1, &tmp);
            if(key &&
               inp.EventType == KEY_EVENT &&
               inp.Event.KeyEvent.bKeyDown) {
                switch(inp.Event.KeyEvent.wVirtualScanCode) {
                case 0x01: /*VK_ESC*/
                    esc_key = TRUE;
                    break;
                case 0x39: /*VK_SPACE*/
                    pause_key = !pause_key;
                    print_log("Log %s\n", !pause_key ? "restarted" : "paused");
                    break;
                case 0x2e: /*VK_C*/
                    if(stdout_mode)
                        break;
                    stdout_copy = !stdout_copy;
                    break;
                case 0x31: /*VK_N*/
                    if(!stdout_mode)
                        next_log = TRUE;
                    break;
                case 0x1f: /*VK_S*/
                    sync_mode = !sync_mode;
                    break;
                case 0x21: /*VK_F*/
                    flush_file = TRUE;
                    break;
                case 0x25: /*VK_K*/
                    skip_kernel = !skip_kernel;
                    print_log("kernel messages %s\n", !skip_kernel ? "on" : "off");
                    break;
                case 0x16: /*VK_Y*/
                    skip_umode = !skip_umode;
                    print_log("user-mode messages %s\n", !skip_umode ? "on" : "off");
                    break;
                }
            }
        }
        Sleep(100);
    }
    return 0;
} // end get_kbd()


/*
 Catch strings from OutputDebugString()
*/
DWORD
__stdcall
get_ODS_messages(PVOID ctx)
{
    HANDLE hCompleteEvent;
    HANDLE hNewMsgEvent;
    HANDLE hdbgf;
    PDBG_OUTPUT_DEBUG_STRING_BUFFER ODSmem;
    ULONG j = 0;
    ULONG status;

    SECURITY_ATTRIBUTES SecAttr;
    SECURITY_DESCRIPTOR SecDesc;

    // Connect to user-mode logger
    memset(&SecAttr, 0, sizeof(SecAttr));
    memset(&SecDesc, 0, sizeof(SecDesc));
    SecAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    SecAttr.bInheritHandle = TRUE;
    SecAttr.lpSecurityDescriptor = &SecDesc;

    if(!InitializeSecurityDescriptor(&SecDesc, SECURITY_DESCRIPTOR_REVISION)) {
        print_log("can't init SecurityDescriptor, (%#x)\n",
            GetLastError());
        return -1;
    }

    if(!SetSecurityDescriptorDacl(&SecDesc, TRUE, NULL, FALSE)) {
        print_log("can't set SD Dacl, (%#x)\n",
            GetLastError());
        return -1;
    }

    hNewMsgEvent = CreateEvent(&SecAttr, FALSE, FALSE, "DBWIN_DATA_READY");

    if(!hNewMsgEvent) {
        print_log("can't create sync event, (%#x)\n",
                GetLastError());
        return -1;
    }

    hCompleteEvent = CreateEvent(&SecAttr, FALSE, FALSE, "DBWIN_BUFFER_READY");

    if(!hCompleteEvent) {
        print_log("can't create sync event (2), (%#x)\n",
                GetLastError());
        return -1;
    }

    hdbgf = CreateFileMapping((HANDLE)(-1),
                        &SecAttr, PAGE_READWRITE,
                        0, 4096,
                        "DBWIN_BUFFER");

    if (!hdbgf) {
        print_log("can't create file mapping, (%#x)\n",
                GetLastError());
        return -1;
    }

    ODSmem = (PDBG_OUTPUT_DEBUG_STRING_BUFFER)MapViewOfFile(hdbgf, FILE_MAP_READ, 0, 0, 512);

    if (!ODSmem) {
        print_log("can't map shared mem (%#x)\n",
                GetLastError());
        return -1;
    }

    // send Ready message
    SetEvent(hCompleteEvent);

    while(!esc_key) {
        // wait for incoming messages
        status = WaitForSingleObject(hNewMsgEvent, INFINITE);
        if(status == WAIT_OBJECT_0) {
            WaitForSingleObject(hODS_lock, INFINITE);

            // copy message to buffer in unified format
            j = MsgCount2a;
            QueryPerformanceCounter((PLARGE_INTEGER)&MsgBuffer2a[j].TimeStamp);
            memcpy(&(MsgBuffer2a[j].Msg[0]), &(ODSmem->Msg[0]), sizeof(ODSmem->Msg));
            MsgBuffer2a[j].Length = strlen((CCHAR*)&(ODSmem->Msg[0]));
            MsgBuffer2a[j].ProcessId = (PVOID)(ODSmem->ProcessId);
            MsgBuffer2a[j].ThreadId = (PVOID)(-1);
            MsgBuffer2a[j].Irql = 0;
            j++;
            MsgCount2a = j;

            ReleaseMutex(hODS_lock);

            SetEvent(hCompleteEvent);
        }
    }
    return 0;
} // end get_ODS_messages()

static const char help_text2[] = 
    "Commands:\n"
    "\t'Esc'   - exit\n"
    "\t'N'     - start new log\n"
    "\t'F'     - flush log buffer\n"
    "Mode switchers (toggle on'off):\n"
    "\t'Space' - pause\n"
    "\t'S'     - synchronous mode\n"
    "\t'C'     - copy log to stdout\n"
    "\t'K'     - capture kernel messages\n"
    "\t'U'     - capture user-mode messages\n"
    ;

/*
 merge user- and kernel-mode logs
*/
void
get_messages()
{
    HANDLE h;
    HANDLE fh;

    ULONG len;
    ULONG i, j, j2;
    ULONG k;
    WCHAR a;
    PWCHAR NumPos;
    ULONG returned;
    BOOLEAN new_line;
    CHAR DbgPrintBuf[1024];
    ULONG SleepCount;
    CHAR tab3[] = "\t\t\t";
    ULONG CurLogSize;
    SYSTEMTIME SysTime;
    LONGLONG FtmSysTime;
    LONGLONG RelSysTime;
    OSVERSIONINFO OsVer;
    PCHAR PlatformStr;

    PDbgPrnHk_GetMessages_USER_OUT tmp_MsgBuffer2;
    DbgPrnHk_GetVersion_USER_OUT VerBuffer;

    BOOLEAN last2 = FALSE;
    PDbgPrnHk_GetMessages_USER_OUT CurMsg = NULL;

    // gain privileges to install/load driver/service
    // it is not necessary here if we are not running as service
    if(sys_account) {
        if(!Privilege(SE_LOAD_DRIVER_NAME, TRUE)) {
            print_log("Warning: Insufficient privileges\n");
        }
    }

    // try to install (or re-install) and start kernel-mode DbgPrint capturing driver
    if(restart_drv) {
        k = NtServiceStop(NT_DbgPrnHk_SVC_NAME);
    }
    k = NtServiceInstall(NT_DbgPrnHk_SVC_NAME, ".");

    k = NtServiceIsRunning(NT_DbgPrnHk_SVC_NAME);
    if(k == 1) {
        // Ok
    } else
    if(k == 0) {
        k = NtServiceStart(NT_DbgPrnHk_SVC_NAME);
    }
    if(k < 0) {
        print_log("Error: Can't load or install DbgPrnHk driver\n");
        exit(-1);
    }

    // Connect to driver
    h = CreateFile(NT_DbgPrnHk_USER_NAME, GENERIC_READ | GENERIC_WRITE,
                   /*FILE_SHARE_READ | FILE_SHARE_WRITE*/ 0,
                   NULL,
                   OPEN_EXISTING,
                   FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
    if(!h || h == (HANDLE)(-1)) {
        print_log("Error: can't open driver\n");
        exit(1);
    }

    // check driver version
    if(!DeviceIoControl(h,IOCTL_DbgPrnHk_GetVersion,
                        NULL,0,
                        &VerBuffer, sizeof(VerBuffer),
                        &returned,NULL)) {
        print_log("Error: incompatible driver version\n");
        exit(1);
    }
    if(VerBuffer.Major != PROD_VER_MJ ||
       VerBuffer.Minor != PROD_VER_MN/* ||
       VerBuffer.Sub   != PROD_VER_NSUB*/) {
        print_log("Error: incompatible driver version\n");
        exit(1);
    }

    if(!stdout_mode) {
        // Init log-file name
        wcsncpy(LogFileN, LogFile, MAX_PATH-10);
        LogFileN[MAX_PATH-10] = 0;
        len = wcslen(LogFileN);

        NumPos = &LogFileN[len];
        LogFileSuf[0] = 0;
        for(i=len; i>0; i--) {
            a = LogFileN[i-1];
            if(a == '\\') {
                break;
            }
            if(a == '.') {
                NumPos = &LogFileN[i-1];
                wcscpy(LogFileSuf, NumPos);
                break;
            }
        }
    }

    // alloc buffers
    MsgBuffer = (PDbgPrnHk_GetMessages_USER_OUT)
        GlobalAlloc(GMEM_DISCARDABLE, BufferSize);
    if(!MsgBuffer) {
        print_log("Error: can't allocate message buffer\n");
        exit(4);
    }
    MsgBuffer2 = (PDbgPrnHk_GetMessages_USER_OUT)
        GlobalAlloc(GMEM_DISCARDABLE, BufferSize);
    if(!MsgBuffer2) {
        print_log("Error: can't allocate message buffer (2)\n");
        exit(4);
    }
    MsgBuffer2a = (PDbgPrnHk_GetMessages_USER_OUT)
        GlobalAlloc(GMEM_DISCARDABLE, BufferSize);
    if(!MsgBuffer2a) {
        print_log("Error: can't allocate message buffer (2)\n");
        exit(4);
    }

    i = Start_i;

    print_log(help_text2);

    if(sys_account) {
        // change service status to RUNNING
        ReportStatusToSCMgr(
            SERVICE_RUNNING,      // service state
            NO_ERROR,             // exit code
            0);                   // wait hint
    }

    // produce logs until break come
    while(!esc_key) {

        if(!stdout_mode) {
            // find unused name for log-file
            if(i>0) {
                swprintf(NumPos, L"%d%s", i, LogFileSuf);
            }
            fh = CreateFileW(LogFileN, GENERIC_READ | GENERIC_WRITE,
                           FILE_SHARE_READ | FILE_SHARE_WRITE,
                           NULL,
                           OPEN_EXISTING,
                           FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
            if(fh && fh != (HANDLE)(-1)) {
                CloseHandle(fh);
                fh = NULL;
                i++;
                if(!i) {
                    print_log("Error: find unused filename\n");
                    exit(2);
                }
                continue;
            }
            if((i >= MaxLogs) && (i-MaxLogs >= Start_i)) {
                if(i-MaxLogs > 0) {
                    swprintf(NumPos, L"%d%s", i-MaxLogs, LogFileSuf);
                } else {
                    swprintf(NumPos, L"%s", LogFileSuf);
                }
                DeleteFileW(LogFileN);
                if(i > 0) {
                    swprintf(NumPos, L"%d%s", i, LogFileSuf);
                } else {
                    swprintf(NumPos, L"%s", LogFileSuf);
                }
            }
            // create log-file
            fh = CreateFileW(LogFileN, GENERIC_READ | GENERIC_WRITE,
                           FILE_SHARE_READ | FILE_SHARE_WRITE,
                           NULL,
                           CREATE_NEW,
                           FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
        } else {
            fh = StdOut;
        }
        if(!fh || fh == (HANDLE)(-1)) {
            print_log("Error: can't create log-file\n");
            fh = NULL;
            exit(3);
        }
        // write log header
        new_line = TRUE;
        if(!stdout_mode) {
            print_log("Writing log to %ws\n", LogFileN);
        }
        CurLogSize = 0;
        len = sprintf(DbgPrintBuf, "Log generated by DbgPrintLog v%d.%d%s\n", PROD_VER_MJ, PROD_VER_MN, PROD_VER_SUB);
        TeeOutput(fh, DbgPrintBuf, len);

        GetSystemTimeAsFileTime((PFILETIME)&FtmSysTime);
        QueryPerformanceCounter((PLARGE_INTEGER)&RelSysTime);
        FileTimeToSystemTime((PFILETIME)&FtmSysTime, &SysTime);
        len = sprintf(DbgPrintBuf, "Timestamp: %d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d (%I64d.%4.4d)\n",
            SysTime.wYear,
            SysTime.wMonth,
            SysTime.wDay,
            SysTime.wHour,
            SysTime.wMinute,
            SysTime.wSecond,
            RelSysTime/10000,
            (ULONG)(RelSysTime%10000)
            );
        TeeOutput(fh, DbgPrintBuf, len);

        OsVer.dwOSVersionInfoSize = sizeof(OsVer);
        GetVersionEx(&OsVer);
        switch(OsVer.dwPlatformId) {
        case VER_PLATFORM_WIN32s:
            PlatformStr = "Win";
            break;
        case VER_PLATFORM_WIN32_WINDOWS:
            PlatformStr = "Win 9x";
            break;
        case VER_PLATFORM_WIN32_NT:
            PlatformStr = "NT";
            break;
        default:
            PlatformStr = "Unknown";
        }
        len = sprintf(DbgPrintBuf, "OS Version: %d:%s %d.%d (%d)\n  %s\n",
            OsVer.dwPlatformId,
            PlatformStr,
            OsVer.dwMajorVersion,
            OsVer.dwMinorVersion,
            OsVer.dwBuildNumber,
            OsVer.szCSDVersion);
        TeeOutput(fh, DbgPrintBuf, len);

        len = sprintf(DbgPrintBuf, "Hostname: ");
        TeeOutput(fh, DbgPrintBuf, len);
        len = 1023;
        GetComputerName(DbgPrintBuf, &len);
        DbgPrintBuf[1023] = 0;
        TeeOutput(fh, DbgPrintBuf, len);

        len = sprintf(DbgPrintBuf, "\nUsername: ");
        TeeOutput(fh, DbgPrintBuf, len);
        len = 1023;
        GetUserName(DbgPrintBuf, &len);
        DbgPrintBuf[1023] = 0;
        TeeOutput(fh, DbgPrintBuf, len);

        TeeOutput(fh, "\n", 1);

        // write to single log-file
        next_log = FALSE;
        while(!esc_key && !next_log) {
            // Get kernel messages
            if(!DeviceIoControl(h,IOCTL_DbgPrnHk_GetMessages,
                                MsgBuffer, min(BufferSize, 0x10000),
                                MsgBuffer, min(BufferSize, 0x10000),
                                &returned,NULL)) {
                returned = GetLastError();
                returned = 0;
            }
            // Get user-mode messages
            WaitForSingleObject(hODS_lock, INFINITE);

            // switch read/write buffers
            tmp_MsgBuffer2 = MsgBuffer2;

            MsgCount2   = MsgCount2a;
            MsgBuffer2  = MsgBuffer2a;

            MsgCount2a  = 0;
            MsgBuffer2a = tmp_MsgBuffer2;

            ReleaseMutex(hODS_lock);

            if(skip_kernel) {
                returned = 0;
            }
            if(skip_umode) {
                MsgCount2 = 0;
            }
            // drop messages if paused
            if(pause_key) {
                returned = 0;
                MsgCount2 = 0;
            }
            if(!returned && !MsgCount2) {
                SleepCount++;
                if(SleepCount > MaxSleepCount) {
                    SleepCount = 0;
                    new_line = TRUE;
                    FlushFileBuffered(fh);
                }
                Sleep(1000);
            }
            SleepCount = 0;
            j = 0;
            j2 = 0;
            // join user- and kernel-mode logs
            while(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT) || MsgCount2) {
                // assure time ordering
                if(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT) && MsgCount2) {
                    if(MsgBuffer[j] .TimeStamp.QuadPart >
                       MsgBuffer2[j2].TimeStamp.QuadPart) {
                        CurMsg = &MsgBuffer2[j2];
                        if(!last2) {
                            new_line = TRUE;
                            last2 = TRUE;
                        }
                    } else {
                        CurMsg = &MsgBuffer[j];
                        if(last2) {
                            new_line = TRUE;
                            last2 = FALSE;
                        }
                    }
                } else
                if(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT)) {
                    CurMsg = &MsgBuffer[j];
                    if(last2) {
                        new_line = TRUE;
                        last2 = FALSE;
                    }
                } else {
                    CurMsg = &MsgBuffer2[j2];
                    if(!last2) {
                        new_line = TRUE;
                        last2 = TRUE;
                    }
                }
                // write message to file (buffered)
                if(CurMsg->Length) {
                    if(new_line) {
                        if(log_mode) {
                            TeeOutput(fh, last2 ? "U " : "K ", 2);
                            CurLogSize+=2;
                        }
                        if(log_irql) {
                            len = sprintf(DbgPrintBuf, "%2.2x ",
                                CurMsg->Irql);
                            TeeOutput(fh, DbgPrintBuf, 3);
                            CurLogSize+=3;
                        }
                        if(log_pid) {
                            len = sprintf(DbgPrintBuf, "%8.8x ",
                                CurMsg->ProcessId);
                            TeeOutput(fh, DbgPrintBuf, len);
                            CurLogSize+=len;
                        }
                        if(log_tid) {
                            if(CurMsg->ThreadId != (PVOID)(-1)) {
                                len = sprintf(DbgPrintBuf, "%8.8x ",
                                    CurMsg->ThreadId);
                            } else {
                                len = sprintf(DbgPrintBuf, "???????? ");
                            }
                            TeeOutput(fh, DbgPrintBuf, len);
                            CurLogSize+=len;
                        }
                        if(!log_time_ex) {
                            len = sprintf(DbgPrintBuf, "%I64d.%4.4d\t",
                                CurMsg->TimeStamp.QuadPart/10000,
                                (ULONG)(CurMsg->TimeStamp.QuadPart%10000));
                        } else {
                            CurMsg->TimeStamp.QuadPart += FtmSysTime - RelSysTime;
                            FileTimeToSystemTime((PFILETIME)&(CurMsg->TimeStamp), &SysTime);
                            len = sprintf(DbgPrintBuf, "[");
                            if(log_time_date) {
                                len += sprintf(DbgPrintBuf+len, "%d/%2.2d/%2.2d ",
                                    SysTime.wYear,
                                    SysTime.wMonth,
                                    SysTime.wDay
                                    );
                            }
                            if(log_time_time) {
                                len += sprintf(DbgPrintBuf+len, "%2.2d:%2.2d:%2.2d.%4.4d ",
                                    SysTime.wHour,
                                    SysTime.wMinute,
                                    SysTime.wSecond,
                                    SysTime.wMilliseconds 
                                    );
                            }
                            if(log_time_rel) {
                                len += sprintf(DbgPrintBuf+len, "(%I64d.%4.4d)",
                                    RelSysTime/10000,
                                    (ULONG)(RelSysTime%10000)
                                    );
                            }
                            len += sprintf(DbgPrintBuf+len, "] ");
                        }
                        TeeOutput(fh, DbgPrintBuf, len);
                        CurLogSize+=len;
                        new_line = FALSE;
                    }
                    len = CurMsg->Length;
                    if(CurMsg->Msg[len-1] == '\n') {
                        new_line = TRUE;
                    }
                    TeeOutput(fh, &CurMsg->Msg[0], len);
                    CurLogSize+=len;
                }
                // go to next message
                if(last2) {
                    j2++;
                    MsgCount2--;
                } else {
                    j++;
                    returned -= sizeof(DbgPrnHk_GetMessages_USER_OUT);
                }
            }
            if(j+j2 < 256) {
                Sleep(100);
            }
            // check user controls
            if(flush_file || sync_mode) {
                if(!stdout_mode && !sync_mode) {
                    print_log("flush log\n");
                }
                FlushFileBuffered(fh);
                flush_file = FALSE;
            }
            if(next_log) {
                print_log("next log\n");
            }
            if(CurLogSize > MaxLogSize && !stdout_mode) {
                next_log = TRUE;
                print_log("max log size limit reached -> next log\n");
            }
        } // end of Get Message loop
        FlushFileBuffered(fh);
        CloseHandle(fh);
        fh = NULL;
    }

    // cleanup
    if(fh) {
        CloseHandle(fh);
    }
    CloseHandle(h);
} // end get_messages()

static const char help_text[] = 
    "Usage:\n"
    "\tDbgPrintLog.exe\t[<switches>] [<log-file name>]\n"
    "Switches:\n"
    "    Log-file format:\n"
    "\t-m       log mode (kernel/user)\n"
    "\t-p       log ProcessId\n"
    "\t-t       log ThreadId (for kernel-mode only)\n"
    "\t-i       log IRQL (for kernel-mode only)\n"
    "\t-full    same as -m -p -t -i\n"
    "\t-t FMT   specify absolute time format FMT. FMT string can contain\n"
    "\t         the following switches: D - date, T - time, R - relative time\n"
//    "\t+fp PID  log only messages with ProcessId==PID\n"
//    "\t-fp PID  do not log messages with ProcessId==PID\n"
    "\t-fm M    do not log messages from M mode (M can be K or U)\n"
    "    Log-file control:\n"
    "\t-s NUM   set max log-file size to NUM Mbytes (0 < NUM < 2048)\n"
    "\t-l NUM   keep NUM latest log files\n"
    "\t-n NUM   set start log number to NUM\n"
    "\t-ft NUM  flush messages buffer after NUM seconds of inactivity\n"
    "\t-sm      synchronous mode. all messages are flushed immediately\n"
    "\t-o\n"
    "\t   or\n"
    "\t--stdout write log to STDOUT istead of file\n"
    "\t-cf      copy log to both STDOUT and file\n"
    "\t-wd DIR  specify working directory\n"
    "    Run mode:\n"
    "\t-svc MOD install or uninstall as service. MOD specify startup mode:\n"
    "\t         A - autimatic, M - manual, U - uninstall. 'A' is used by default\n"
    "\t         if MOD is omited.\n"
    "\t-drv MOD Specify driver startup mode:\n"
    "\t         B - boot, S - system, A - autimatic, M - manual. 'M' is used by default\n"
    "\t         if MOD is omited.\n"
    "\t-rd      restart and reinstall driver before start\n"
    "\t-h       display this help message\n"
    ;

void
usage(void)
{
    HANDLE stdOuth = GetStdHandle(STD_OUTPUT_HANDLE);
    ULONG WrittenBytes;

    if(stdOuth && stdOuth != ((HANDLE)(-1))) {
        WriteFile(stdOuth, help_text,  strlen(help_text),  &WrittenBytes, NULL);
        WriteFile(stdOuth, help_text2, strlen(help_text2), &WrittenBytes, NULL);
    } else {
        MessageBox(NULL, help_text, "Usage", MB_OK);
    }
} // end usage()

BOOL
ReportStatusToSCMgr(
    DWORD dwCurrentState,
    DWORD dwWin32ExitCode,
    DWORD dwWaitHint
    )
{
    static DWORD dwCheckPoint = 1;
    BOOL fResult = TRUE;

    if (dwCurrentState == SERVICE_START_PENDING) {
        ssStatus.dwControlsAccepted = 0;
    } else {
        ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_PAUSE_CONTINUE;
    }

    ssStatus.dwCurrentState = dwCurrentState;
    ssStatus.dwWin32ExitCode = dwWin32ExitCode;
    ssStatus.dwWaitHint = dwWaitHint;

    if ( ( dwCurrentState == SERVICE_RUNNING ) ||
         ( dwCurrentState == SERVICE_STOPPED ) ) {
        ssStatus.dwCheckPoint = 0;
    } else {
        ssStatus.dwCheckPoint = dwCheckPoint++;
    }

    // Report the status of the service to the service control manager.
    if (!(fResult = SetServiceStatus( sshStatusHandle, &ssStatus))) {
        AddToMessageLog(TEXT("SetServiceStatus"));
    }
    return fResult;
} // end ReportStatusToSCMgr()

VOID
AddToMessageLog(
    LPTSTR lpszMsg
    )
{
    CHAR    szMsg[256];
    HANDLE  hEventSource;
    LPSTR   lpszStrings[2];
    ULONG   dwErr;

    dwErr = GetLastError();

    // Use event logging to log the error.
    hEventSource = RegisterEventSource(NULL, NT_DbgPrnHk_Client_SVC_NAME);

    sprintf(szMsg, "%s error: %d", NT_DbgPrnHk_Client_SVC_NAME, dwErr);
    lpszStrings[0] = szMsg;
    lpszStrings[1] = lpszMsg;

    if (!hEventSource)
        return;

    ReportEvent(hEventSource, // handle of event source
        EVENTLOG_ERROR_TYPE,  // event type
        0,                    // event category
        0,                    // event ID
        NULL,                 // current user's SID
        2,                    // strings in lpszStrings
        0,                    // no bytes of raw data
        (const char **)lpszStrings,          // array of error strings
        NULL);                // no raw data

    (VOID)DeregisterEventSource(hEventSource);
} // end AddToMessageLog()

void
WINAPI
service_ctrl(
    DWORD dwCtrlCode)
{
    ULONG sst;
    // Handle the requested control code.
    switch(dwCtrlCode)
    {
    // Stop the service.

    // SERVICE_STOP_PENDING should be reported before
    // setting the Stop Event - hServerStopEvent - in
    // ServiceStop().  This avoids a race condition
    // which may result in a 1053 - The Service did not respond...
    // error.
    case SERVICE_CONTROL_STOP:
        ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 3000);
        esc_key = TRUE;
        break;

    case SERVICE_CONTROL_PAUSE:
        ReportStatusToSCMgr(SERVICE_PAUSED, NO_ERROR, 0);
        pause_key = TRUE;
        break;

    case SERVICE_CONTROL_CONTINUE:
        ReportStatusToSCMgr(SERVICE_RUNNING, NO_ERROR, 0);
        pause_key = FALSE;
        break;

    default:
    case SERVICE_CONTROL_INTERROGATE:
        if(esc_key) {
            sst = SERVICE_STOP_PENDING;
        } else
        if(pause_key) {
            sst = SERVICE_PAUSED;
        } else {
            sst = SERVICE_RUNNING;
        }
        ReportStatusToSCMgr(sst, NO_ERROR, 0);
        break;
    }
} // service_ctrl()

void
WINAPI
service_main(
    DWORD dwArgc,
    LPTSTR *lpszArgv)
{
    // register our service control handler:
    sshStatusHandle = RegisterServiceCtrlHandler( NT_DbgPrnHk_Client_SVC_NAME, service_ctrl);
    print_log("***** DbgPrintLog service start********");
    if (!sshStatusHandle)
        goto cleanup;

    // SERVICE_STATUS members that don't change in example
    ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    ssStatus.dwServiceSpecificExitCode = 0;


    // report the status to the service control manager.
    if (!ReportStatusToSCMgr(
        SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        3000))                 // wait hint
        goto cleanup;

    // get messages from user- and kernel-mode and store them to log-file
    get_messages();

cleanup:

    // try to report the stopped status to the service control manager.
    if (sshStatusHandle)
        (VOID)ReportStatusToSCMgr(
                            SERVICE_STOPPED,
                            NO_ERROR,
                            0);

    return;
} // end service_main()

void main () {
    int argc;
    WCHAR** argv;
    WCHAR*  CmdLine;
    WCHAR*  CmdLine2;

    LONG i=0;
    LONG j;
    ULONG k, t = REG_SZ;
    WCHAR a;
    ULONG thId;
    HKEY hKey;
    WCHAR WDir[MAX_PATH*2];

    print_log("DbgPrintLog v%d.%d%s (c) by Alter\nHome site http://www.alter.org.ua\n",
        PROD_VER_MJ, PROD_VER_MN, PROD_VER_SUB);
    fflush(stdout);

    if(GetVersion() & 0x80000000) {
        print_log("For Windows NT family only.\n");
        exit(-2);
    }

//    __asm int 3;
    sys_account = IsSystemAccount();

    WDir[0] = 0;
    GetCurrentDirectoryW(MAX_PATH*2, WDir);
    WDir[MAX_PATH*2-1] = 0;

    if(sys_account) {
        k = MAX_PATH*3;
        CmdLine = (PWCHAR)GlobalAlloc(GMEM_DISCARDABLE, k);
        if(!CmdLine) {
            exit(-3);
        }
        if(RegOpenKey(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\" NT_DbgPrnHk_Client_SVC_NAME, &hKey) != ERROR_SUCCESS) {
            exit(-4);
        }
        if(RegQueryValueExW(hKey, NT_DbgPrnHk_Reg_CmdLineW, NULL, &t, (PUCHAR)CmdLine, &k) != ERROR_SUCCESS) {
            GlobalFree(CmdLine);
            CmdLine = GetCommandLineW();
        }
        k = MAX_PATH*2;
        if(RegQueryValueExW(hKey, NT_DbgPrnHk_Reg_WDirW, NULL, &t, (PUCHAR)WDir, &k) != ERROR_SUCCESS) {
            // do nothing
        }
        RegCloseKey(hKey);
    } else {
        CmdLine = GetCommandLineW();
        if(!CmdLine) {
            exit(-1);
        }
    }
    argv = CommandLineToArgvW(CmdLine, &argc);

    for(i=1; i<argc; i++) {
        if(!argv[i] || argv[i][0] != '-') {
            LogFile = argv[i];
            continue;
        }
        if(!wcscmp(argv[i], L"-h") ||
           !wcscmp(argv[i], L"--help")) {
            usage();
            exit(0);
        } else
        if(!wcscmp(argv[i], L"-sm")) {
            sync_mode = TRUE;
        } else
        if(!wcscmp(argv[i], L"-wd")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-1);
            }
            wcsncpy(WDir, argv[i], MAX_PATH*2-1);
            WDir[MAX_PATH*2-1] = 0;
        } else
        if(!wcscmp(argv[i], L"-drv")) {
            i++;
            if(i < argc) {
                if(!wcscmp(argv[i], L"B")) {
                    drv_mode = SERVICE_BOOT_START;
                } else
                if(!wcscmp(argv[i], L"S")) {
                    drv_mode = SERVICE_SYSTEM_START;
                } else
                if(!wcscmp(argv[i], L"A")) {
                    drv_mode = SERVICE_AUTO_START;
                } else
                if(!wcscmp(argv[i], L"M")) {
                    drv_mode = SERVICE_DEMAND_START;
                } else
                if(!wcscmp(argv[i], L"")) {
                    if(sys_account) {
                        continue;
                    }
                    k = NtServiceRemove(NT_DbgPrnHk_SVC_NAME);
                    if(k == 1) {
                        print_log("DbgPrnHk driver is uninstalled\n");
                    }
                    deinstall = TRUE;
                    continue;
                } else {
                    // revert to initial state
                    i--;
                }
            }
            if(sys_account) {
                continue;
            }
            install_drv = TRUE;
        } else
        if(!wcscmp(argv[i], L"-svc")) {
            i++;
            if(i < argc) {
                if(!wcscmp(argv[i], L"U")) {
                    if(sys_account) {
                        continue;
                    }
                    k = NtServiceRemove(NT_DbgPrnHk_Client_SVC_NAME);
                    if(k == 1) {
                        print_log("DbgPrintLog service is uninstalled\n");
                    }
                    deinstall = TRUE;
                    continue;
                } else
                if(!wcscmp(argv[i], L"A")) {
                    cli_mode = SERVICE_AUTO_START;
                } else
                if(!wcscmp(argv[i], L"M")) {
                    cli_mode = SERVICE_DEMAND_START;
                } else {
                    // revert to initial state
                    i--;
                }
            }
            if(sys_account) {
                continue;
            }
            install_svc = TRUE;
        } else
        if(!wcscmp(argv[i], L"-s")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            swscanf(argv[i], L"%d", &MaxLogSize);
            if(!MaxLogSize || MaxLogSize > 2047) {
                usage();
                exit(-2);
            }
            MaxLogSize = MaxLogSize << 20;
        } else
        if(!wcscmp(argv[i], L"-l")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            swscanf(argv[i], L"%d", &MaxLogs);
        } else
        if(!wcscmp(argv[i], L"-n")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            swscanf(argv[i], L"%d", &Start_i);
        } else
        if(!wcscmp(argv[i], L"-full")) {
            log_irql = TRUE;
            log_pid = TRUE;
            log_tid = TRUE;
            log_mode = TRUE;
        } else
        if(!wcscmp(argv[i], L"-ft")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            swscanf(argv[i], L"%d", &MaxSleepCount);
        } else
        if(!wcscmp(argv[i], L"-fm")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            if(!wcsicmp(argv[i], L"K")) {
                skip_kernel = TRUE;
            } else
            if(!wcsicmp(argv[i], L"U")) {
                skip_umode = TRUE;
            } else {
                usage();
                exit(-2);
            }
        } else
        if(!wcscmp(argv[i], L"-t")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            for(j=0; a = argv[i][j]; j++) {
                switch(a) {
                case 'D':
                case 'd':
                    log_time_date = TRUE;
                    break;
                case 'T':
                case 't':
                    log_time_time = TRUE;
                    break;
                case 'R':
                case 'r':
                    log_time_rel = TRUE;
                    break;
                default:
                    usage();
                    exit(-2);
                }
                log_time_ex = TRUE;
            }
        } else
        if(!wcscmp(argv[i], L"-cf")) {
            stdout_copy = TRUE;
            stdout_mode = FALSE;
        } else
        if(!wcscmp(argv[i], L"--stdout") ||
           !wcscmp(argv[i], L"-o")) {
            stdout_mode = TRUE;
            stdout_copy = FALSE;
        } else
        if(!wcscmp(argv[i], L"-i")) {
            log_irql = TRUE;
        } else
        if(!wcscmp(argv[i], L"-p")) {
            log_pid = TRUE;
        } else
        if(!wcscmp(argv[i], L"-t")) {
            log_tid = TRUE;
        } else
        if(!wcscmp(argv[i], L"-m")) {
            log_mode = TRUE;
        } else
        if(!wcscmp(argv[i], L"-rd")) {
            restart_drv = TRUE;
        }
    }

    StdIn  = GetStdHandle(STD_INPUT_HANDLE);
    StdOut = GetStdHandle(STD_OUTPUT_HANDLE);

    // gain privileges to install/load driver/service
    if(!Privilege(SE_LOAD_DRIVER_NAME, TRUE)) {
        print_log("Warning: Insufficient privileges\n");
    }

    // do nothing if this was service/driver deinstall
    if(deinstall) {
        exit(0);
    }

    // set current directory
    SetCurrentDirectoryW(WDir);

    if(install_drv) {
        k = NtServiceRemove(NT_DbgPrnHk_SVC_NAME);
        k = NtServiceInstall(NT_DbgPrnHk_SVC_NAME, ".", TRUE, drv_mode);
        if(k < 0) {
            print_log("Error: Can't install DbgPrnHk driver\n");
            exit(-1);
        }
    }
    if(install_svc) {
        // try to install (or re-install) DbgPrint capturing service
        k = NtServiceRemove(NT_DbgPrnHk_Client_SVC_NAME);
        k = NtServiceInstall(NT_DbgPrnHk_Client_SVC_NAME, ".", FALSE, cli_mode);
        if(k < 0) {
            print_log("Error: Can't install DbgPrintLog service\n");
            exit(-1);
        }
        // save command line in registry
        if(RegOpenKey(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\" NT_DbgPrnHk_Client_SVC_NAME, &hKey) != ERROR_SUCCESS) {
            print_log("Error: Can't store parameters in registry\n");
            exit(-4);
        }
        CmdLine2 = GetCommandLineW();
        if(!CmdLine2) {
            exit(-1);
        }
        k = (wcslen(CmdLine2)+1)*sizeof(WCHAR);
        if(RegSetValueExW(hKey, NT_DbgPrnHk_Reg_CmdLineW, NULL, REG_SZ, (PUCHAR)CmdLine2, k) != ERROR_SUCCESS) {
            print_log("Error: Can't store parameters in registry\n");
            exit(-4);
        }
        k = (wcslen(WDir)+1)*sizeof(WCHAR);
        if(RegSetValueExW(hKey, NT_DbgPrnHk_Reg_WDirW, NULL, REG_SZ, (PUCHAR)WDir, k) != ERROR_SUCCESS) {
            print_log("Error: Can't store parameters in registry\n");
            exit(-4);
        }
        RegCloseKey(hKey);
        // ensure that driver is installed
        k = NtServiceInstall(NT_DbgPrnHk_SVC_NAME, ".", TRUE, drv_mode);

        print_log("DbgPrintLog service is installed\n");
    }

    if(install_drv || install_svc) {
        print_log("Changes will take effect after reboot\n");
        print_log("Use 'net start/stop DbgPrintLog' to start/stop logging immediately\n");
        exit(0);
    }

    // init lock for switching buffers
    hODS_lock = CreateMutex(NULL, FALSE, NULL);
    if(!hODS_lock || hODS_lock == (HANDLE)(-1)) {
        print_log("can't init mutex, (%#x)\n",
            GetLastError());
        exit(-1);
    }

    // start user-mode OutputDebugString capturing thread
    CreateThread(NULL, 0, get_ODS_messages, NULL, 0, &thId);
    if(sys_account) {
        // init as service
        if (!StartServiceCtrlDispatcher(dispatchTable)) {
            AddToMessageLog("StartServiceCtrlDispatcher failed.");
            exit(-1);
        }
    } else {
        // start keyboard event handling thread
        CreateThread(NULL, 0, get_kbd, NULL, 0, &thId);
        // get messages from user- and kernel-mode and store them to log-file
        get_messages();
    }

    exit(0);
} // end main()

